#if defined (WIN32) && defined (UNICODE)
#include <windows.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "pb.h"

#define FFSOX_BUF_SIZE 1024
#define FOLDER_CSV "folder.csv"

///////////////////////////////////////////////////////////////////////////////
typedef struct priv {
  char *buf;
  char *wp;
  char *mp;
  FILE *f;
  char ch;
} priv_t;

static priv_t *priv_init(priv_t *b, FILE *f);
static priv_t *priv_cleanup(priv_t *b);
static int priv_realloc(priv_t *b, size_t n);
static int priv_gets(priv_t *b);

///////////////////////////////////////////////////////////////////////////////
static priv_t *priv_init(priv_t *b, FILE *f)
{
  memset(b,0,sizeof *b);

  if (NULL==(b->buf=malloc(FFSOX_BUF_SIZE)))
    goto error;

  b->wp=b->buf;
  b->mp=b->buf+FFSOX_BUF_SIZE;
  b->f=f;
  b->ch=fgetc(f);

  return b;
error:
  priv_cleanup(b);

  return NULL;
}

static priv_t *priv_cleanup(priv_t *b)
{
  if (NULL!=b->buf)
    free(b->buf);

  return b;
}

static int priv_realloc(priv_t *b, size_t n)
{
  size_t size=b->mp-b->buf;
  size_t offs=b->wp-b->buf;
  char *tmp;

  do {
    size*=2;
  } while (size<offs+n);

  if (NULL==(tmp=realloc(b->buf,size)))
    return -1;

  b->buf=tmp;
  b->wp=b->buf+offs;
  b->mp=b->buf+size;

  return 0;
}

static int priv_gets(priv_t *b)
{
  char ch=b->ch;
#if defined (WIN32) && defined (UNICODE)
  int i,n;
  char s[8];
  wchar_t ws[8];

  enum {
    NCH=sizeof s/sizeof s[0],
    NWCH=sizeof ws/sizeof ws[0]
  };
#endif

  b->wp=b->buf;

  for (;;) {
    if (EOF==ch)
      goto end_line;
    else if ('\r'==ch) {
      if ('\n'==(ch=fgetc(b->f)))
        ch=fgetc(b->f);

      goto end_line;
    }
    else if ('\n'==ch) {
      if ('\r'==(ch=fgetc(b->f)))
        ch=fgetc(b->f);

      goto end_line;
    }

#if defined (WIN32) && defined (UNICODE)
    // convert ANSI to wide char
    n=MultiByteToWideChar(
      CP_ACP,   // __in       UINT CodePage,
      0,        // __in       DWORD dwFlags,
      &ch,      // __in       LPCSTR lpMultiByteStr,
      1,        // __in       int cbMultiByte,
      ws,       // __out_opt  LPWSTR lpWideCharStr,
      NWCH      // __in       int cchWideChar
    );

    if (n<1||NWCH<n)
      return -1;

    // convert wide char to UTF-8
    n=WideCharToMultiByte(
      CP_UTF8,  // __in       UINT CodePage
      0,        // __in       DWORD dwFlags
      ws,       // __in       LPCWSTR lpWideCharStr
      n,        // __in       int cchWideChar
      s,        // __out_opt  LPSTR lpMultiByteStr
      NCH,      // __in       int cbMultiByte
      NULL,     // __in_opt   LPCSTR lpDefaultChar
      NULL      // __out_opt  LPBOOL lpUsedDefaultChar
    );

    if (n<1||NCH<n)
      return -1;
    else if (b->mp<b->wp+n&&0!=priv_realloc(b,n))
      return -1;

    for (i=0;i<n;++i)
      *b->wp++=s[i];
#else
    if (b->wp==b->mp&&0!=priv_realloc(b,1))
      return -1;

    *b->wp++=ch;
#endif

    ch=fgetc(b->f);

    continue;
  end_line:
    if (b->wp==b->mp&&0!=priv_realloc(b,1))
      return -1;

    b->ch=ch;
    *b->wp++=0;

    break;
  }

  return 0;
}

#if defined (WIN32) && defined (UNICODE)
static FILE *priv_fopen(const char *name, const wchar_t *wmode)
{
  int n1,n2;
  wchar_t *wname=NULL;
  FILE *f=NULL;

  n1=MultiByteToWideChar(
    CP_UTF8,      // __in       UINT CodePage,
    0,            // __in       DWORD dwFlags,
    name,         // __in       LPCSTR lpMultiByteStr,
    -1,           // __in       int cbMultiByte,
    NULL,         // __out_opt  LPWSTR lpWideCharStr,
    0             // __in       int cchWideChar
  );

  if (0==n1||NULL==(wname=malloc(n1*sizeof *wname)))
	goto cleanup;

  n2=MultiByteToWideChar(
    CP_UTF8,      // __in       UINT CodePage,
    0,            // __in       DWORD dwFlags,
    name,         // __in       LPCSTR lpMultiByteStr,
    -1,           // __in       int cbMultiByte,
    wname,        // __out_opt  LPWSTR lpWideCharStr,
    n1            // __in       int cchWideChar
  );

  if (n2!=n1)
	goto cleanup;

  f=_wfopen(wname,wmode);
cleanup:
  if (NULL!=wname)
	free(wname);

  return f;
}
#endif

int pb_csv2avdict(char *file, char sep, pb_dict_set_t dict_set,
    struct AVDictionary **dict)
{
  char *name=pb_name(file);
  char *csv=NULL;
  FILE *f=NULL;
  priv_t b;
  char tok[]={sep,0};
  char *head=NULL;
  char *rp=NULL;
  char *pp=NULL;
  char *hrp=NULL;
  char *hmp;
#if ! defined (WIN32)
  char *np=NULL;
#endif
  int code=-1;

  memset(&b,0,sizeof b);

  if (NULL==(csv=malloc((name-file)+strlen(FOLDER_CSV)+1)))
    goto cleanup;

  memcpy(csv,file,name-file);
  strcpy(csv+(name-file),FOLDER_CSV);

#if defined (WIN32) && defined (UNICODE)
  if (NULL==(f=priv_fopen(csv,L"rb")))
    goto cleanup;
#else
  if (NULL==(f=fopen(csv,"rb")))
    goto cleanup;
#endif

  // initilialize parser.
  if (NULL==priv_init(&b,f))
    goto cleanup;

  // read the header.
  if (0!=priv_gets(&b))
    goto cleanup;

  if (NULL==(head=malloc(b.wp-b.buf)))
    goto cleanup;

  memcpy(head,b.buf,b.wp-b.buf);
  hmp=head+(b.wp-b.buf);

#if defined (WIN32)
  for (rp=strtok(head,tok);NULL!=rp;rp=strtok(NULL,tok))
#else
  for (rp=strtok_r(head,tok,&np);NULL!=rp;rp=strtok_r(NULL,tok,&np))
#endif
    ;

  if (0!=strcmp("file",head))
    goto cleanup;

  do {
    if (0!=priv_gets(&b))
      goto cleanup;

    pp=b.buf;
#if defined (WIN32)
    rp=strtok(b.buf,tok);
#else
    rp=strtok_r(b.buf,tok,&np);
#endif

    for (hrp=head;hrp<hmp&&NULL!=hrp&&NULL!=rp;hrp=hrp+strlen(hrp)+1) {
      if (NULL==rp||pp<rp)
        ++pp;
      else {
        if (0==strcmp("file",hrp)||0==*rp) {
          if (0!=strcmp(name,rp))
            goto next_line;
        }
        else if (dict_set(dict,hrp,rp,0)<0)
          goto cleanup;

        pp=rp+strlen(rp)+1;
#if defined (WIN32)
        rp=strtok(NULL,tok);
#else
        rp=strtok_r(NULL,tok,&np);
#endif
      }
    }

    code=0;
    break;
  next_line:
    ;
  } while (EOF!=b.ch);
cleanup:
  if (NULL!=head)
    free(head);

  priv_cleanup(&b);

  if (NULL!=f)
    fclose(f);

  if (NULL!=csv)
    free(csv);

  return code;
}

#if defined (__TEST__)
///////////////////////////////////////////////////////////////////////////////
#define FN "Joe_Walsh_-_Rocky_Mountain_Way_-_Vintage_Live_Footage_1972.flv"
//#define FN "Joe_Walsh_-_Turn_To_Stone_-_Vintage_Live_Footage_1972.flv"

int main(int argc, char **argv)
{
  AVDictionary *dict=NULL;
  AVDictionaryEntry *tag=NULL;

  if (0!=priv_csv2dict(argv[1],FN,'\t',&dict))
    goto cleanup;

  while (NULL!=(tag=av_dict_get(dict,"",tag,AV_DICT_IGNORE_SUFFIX)))
    printf("%s=%s\n",tag->key,tag->value);
cleanup:
  av_dict_free(&dict);

  return 0;
}

int main1(int argc, char **argv)
{
  FILE *f=NULL;
  priv_t b;
  char *head=NULL;
  char *rp=NULL;
  char *pp=NULL;
  char *hrp=NULL;
  char *hmp;
#if ! defined (WIN32)
  char *np=NULL;
#endif

  memset(&b,0,sizeof b);

  // open input file.
  if (NULL==(f=fopen(argv[1],"rb")))
    goto cleanup;

  // initilialize parser.
  if (NULL==priv_init(&b,f))
    goto cleanup;

  // read the header.
  if (0!=priv_gets(&b))
    goto cleanup;

  if (NULL==(head=malloc(b.wp-b.buf)))
    goto cleanup;

  memcpy(head,b.buf,b.wp-b.buf);
  hmp=head+(b.wp-b.buf);

#if defined (WIN32)
  for (rp=strtok(head,"\t");NULL!=rp;rp=strtok(NULL,"\t"))
#else
  for (rp=strtok_r(head,"\t",&np);NULL!=rp;rp=strtok_r(NULL,"\t",&np))
#endif
    //printf("%s\n",rp);
    ;

  do {
    if (0!=priv_gets(&b))
      goto cleanup;

    pp=b.buf;
    rp=strtok(b.buf,"\t");

    for (hrp=head;hrp<hmp&&NULL!=hrp&&NULL!=rp;hrp=hrp+strlen(hrp)+1) {
      if (NULL==rp||pp<rp)
        ++pp;
      else {
        printf("%s: %s\n",hrp,rp);
        pp=rp+strlen(rp)+1;
        rp=strtok(NULL,"\t");
      }
    }

    printf("\n");
  } while (EOF!=b.ch);
cleanup:
  if (NULL!=head)
    free(head);

  priv_cleanup(&b);

  if (NULL!=f)
    fclose(f);

  return 0;
}
#endif
